@@ -1,4 +1,4 @@
-# (c) 2003-2012 Vlado Keselj http://web.cs.dal.ca/~vlado
+# (c) 2003-2017 Vlado Keselj http://web.cs.dal.ca/~vlado
 #
 # Text::Ngrams - A Perl module for N-grams processing
 
@@ -12,11 +12,10 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
 %EXPORT_TAGS = ( 'all' => [ qw(new encode_S decode_S) ] );
 @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
 @EXPORT = qw(new);
-$VERSION = '2.003';
+$VERSION = '2.005';
 
-use vars qw($Version $Revision);
+use vars qw($Version);
 $Version = $VERSION;
-($Revision = substr(q$Revision: 395 $, 10)) =~ s/\s+$//;
 
 use vars @EXPORT_OK;
 use vars qw();			# non-exported package globals go here
@@ -71,8 +70,10 @@ sub new {
       #so that for the output the n-grams are in utf8 (encoded by encode_S though)
       
       $self->{allow_iproc} = 0;
-      #allow_iproc has to be 0. Otherwise the last token in the read block will be preprocessed and encoded in utf8,
-      #and then attached at the beginning of the next block read from input, which will be in the internal Perl representation
+      # allow_iproc has to be 0. Otherwise the last token in the read block will
+      # be preprocessed and encoded in utf8,
+      # and then attached at the beginning of the next block read from input,
+      # which will be in the internal Perl representation
   }
   #MJ <-
   elsif ($params{type} eq 'byte') {
@@ -306,6 +307,7 @@ sub _reduce_to_limit {
 		die if $nextprunefrequency <= $prunefrequency;
 }   }   }   }
 
+# Sorts keys according to the lexicographic order.
 sub _keys_sorted {
     my $self = shift;
     my $n = shift;
@@ -330,17 +332,19 @@ sub get_ngrams {
     my $onlyfirst = exists($params{'onlyfirst'}) ? $params{'onlyfirst'} : '';
     my $opt_normalize = exists($params{'normalize'}) ?$params{'normalize'} : '';
 
-    my $total = $self->{total}[$n];
-    my @keys;
-    if (!exists($params{'orderby'}) or $params{'orderby'} eq 'ngram')
-    { @keys = $self->_keys_sorted($n) }
-    elsif ($params{'orderby'} eq 'none') {
-        die "onlyfirst requires order" if $onlyfirst;
-        @keys = keys(%{$self->{table}[$n]})
-	}
+    my $total = $self->{total}[$n]; my @keys = ();
+    if (!exists($params{'orderby'}) or $params{'orderby'} eq 'ngram') {
+      @keys = $self->_keys_sorted($n);
+    } elsif ($params{'orderby'} eq 'none') {
+      die "onlyfirst requires order" if $onlyfirst;
+      @keys = keys(%{$self->{table}[$n]})
+    }
     elsif ($params{'orderby'} eq 'frequency') {
-        @keys = sort { $self->{table}[$n]{$b} <=>
-		       $self->{table}[$n]{$a} }
+      @keys = $self->_keys_sorted($n);
+      my %keysord = ();
+      for (my $i=0; $i<=$#keys; ++$i) { $keysord{$keys[$i]} = $i }
+      @keys = sort { $self->{table}[$n]{$b} <=> $self->{table}[$n]{$a}
+		   or $keysord{$a} <=> $keysord{$b} }
 	keys(%{$self->{table}[$n]});
     }
     else { die }
@@ -360,10 +364,13 @@ sub get_ngrams {
 sub to_string {
     my $self = shift;
     my (%params) = @_;
+    my $n = exists($params{'n'})? $params{'n'} : $self->{windowsize};
+    my $onlyfirst = exists($params{'onlyfirst'}) ? $params{'onlyfirst'} : '';
+    my $opt_normalize = exists($params{'normalize'}) ?$params{'normalize'} : '';
    
-    my $onlyfirst = exists($params{'onlyfirst'}) ?
-	$params{'onlyfirst'} : '';
-    delete $params{'onlyfirst'};
+    #my $onlyfirst = exists($params{'onlyfirst'}) ?
+    #$params{'onlyfirst'} : '';
+    #delete $params{'onlyfirst'};
 
     my $out =  exists($params{'out'}) ? $params{'out'} : '';
     delete $params{'out'};
@@ -373,8 +380,8 @@ sub to_string {
         $outh = *FH;
     }
 
-    my $opt_normalize = $params{'normalize'};
-    delete $params{'normalize'};
+    #my $opt_normalize = $params{'normalize'};
+    #delete $params{'normalize'};
 
     my $spartan = $params{'spartan'};
     delete $params{'spartan'};
@@ -409,13 +416,13 @@ sub to_string {
 
 	@keys = splice(@keys,0,$onlyfirst) if $onlyfirst;
 
-	foreach my $ngram (@keys) {
-	    my $count = $self->{table}[$n]{$ngram};
-	    $ret .= $self->_encode_S($ngram) . "\t" .
-		($opt_normalize ? ($count / $total ) : $count) ."\n";
-	    if ($out) { print $outh $ret; $ret = '' }
-
+	my %params1=%params; $params1{n}=$n;
+	my @a = $self->get_ngrams(%params1);
+	for (my $i=0; $i<=$#a; $i+=2) {
+	  my $ng = $a[$i]; my $f = $a[$i+1];
+	  $ret.="$ng\t$f\n";
 	}
+	if ($out) { print $outh $ret; $ret = '' }
 
 	$ret .= "\n" unless $spartan;
     }
@@ -430,7 +437,7 @@ sub to_string {
     return $ret;
 }
 
-# http://www.cs.dal.ca/~vlado/srcperl/snip/decode_S
+# http://web.cs.dal.ca/~vlado/srcperl/snip/decode_S
 sub decode_S ( $ ) {
     local $_ = shift;
     my $out;
@@ -502,10 +509,12 @@ Text::Ngrams - Flexible Ngram analysis (for characters, words, and more)
 For default character n-gram analysis of string:
 
   use Text::Ngrams;
-  my $ng3 = Text::Ngrams->new;
-  $ng3->process_text('abcdefg1235678hijklmnop');
+  my $ng = Text::Ngrams->new;
+  $ng->process_text('abcdefg1235678hijklmnop');
   print $ng3->to_string;
-  my @ngramsarray = $ng3->get_ngrams;
+  my @ngramsarray = $ng->get_ngrams;
+  # or put ngrams and frequencies into a hash
+  my %ngrams = $ng3->get_ngrams( n => 3, normalize => 1 );
 
 One can also feed tokens manually:
 
@@ -513,12 +522,8 @@ One can also feed tokens manually:
   my $ng3 = Text::Ngrams->new;
   $ng3->feed_tokens('a');
   $ng3->feed_tokens('b');
-  $ng3->feed_tokens('c');
-  $ng3->feed_tokens('d');
-  $ng3->feed_tokens('e');
-  $ng3->feed_tokens('f');
-  $ng3->feed_tokens('g');
-  $ng3->feed_tokens('h');
+  $ng3->feed_tokens('c', 'd');
+  $ng3->feed_tokens(qw(e f g h));
 
 We can choose n-grams of various sizes, e.g.:
 
@@ -568,7 +573,7 @@ C<ngrams.pl> provided with the package.
 
 The output looks like this (version number may be different):
 
-  BEGIN OUTPUT BY Text::Ngrams version 1.1
+  BEGIN OUTPUT BY Text::Ngrams version 2.004
 
   1-GRAMS (total count: 8)
   ------------------------
@@ -603,7 +608,7 @@ The output looks like this (version number may be different):
   END OUTPUT BY Text::Ngrams
 
 N-grams are encoded using encode_S
-(F<www.cs.dal.ca/~vlado/srcperl/snip/encode_S>), so that they can
+(F<web.cs.dal.ca/~vlado/srcperl/snip/encode_S>), so that they can
 always be recognized as \S+.  This encoding does not change strings
 "too much", e.g., letters, digits, and most punctuation characters
 will remail unchanged, and space is replaced by underscore (_).
@@ -614,7 +619,7 @@ and vice versa.
 
 An example of word n-grams containing space:
 
-  BEGIN OUTPUT BY Text::Ngrams version 1.1
+  BEGIN OUTPUT BY Text::Ngrams version 2.004
 
   1-GRAMS (total count: 8)
   ------------------------
@@ -635,7 +640,7 @@ An example of word n-grams containing space:
 
 Or, in case of byte type of processing:
 
-  BEGIN OUTPUT BY Text::Ngrams version 1.1
+  BEGIN OUTPUT BY Text::Ngrams version 2.004
 
   1-GRAMS (total count: 55)
   -------------------------
@@ -673,7 +678,7 @@ Or, in case of byte type of processing:
 
 =head1 METHODS
 
-=head2 new ( windowsize => POS_INTEGER, type => character|byte|word|utf8|utf8_character, limit => POS_INTEGER )
+=head2 new ( windowsize => POS_INTEGER, type => 'character' | 'byte' | 'word' | 'utf8' | 'utf8_character', limit => POS_INTEGER )
 
   my $ng = Text::Ngrams->new;
   my $ng = Text::Ngrams->new( windowsize=>10 );
@@ -688,10 +693,11 @@ Parameters:
 
 =item limit
 
-Limit the number of distinct n-grams.  Processing large files may be
+Limit the number of distinct n-grams collected during processing.  Processing large files may be
 slow, so you can limit the total number of distinct n-grams which are
-counted to speed up processing.
-Less-frequent ones will be deleted.
+counted to speed up processing.  The speed-up is implemented by periodically prunning the
+collected n-gram.  Due to this process, the final n-gram counts may not be correct, and
+the list of final most frequen n-grams may not be correct either.
 
 B<BEWARE:> If a limit is set, the n-gram counts at the end may not be
 correct due to periodical pruning of n-grams.
@@ -730,6 +736,7 @@ are replaced by <NUMBER>, and everything else is ignored.  A space is inserted
 when n-grams are formed.       
 
 =item utf8_character
+
 UTF8 analogue of the "character" type: from a UTF8 encoded text reads letters,
 sequences of all other characters are replaced by a space, letters are turned uppercase 
 
@@ -773,17 +780,20 @@ foolowing way:
   }
   elsif ($params{type} eq 'utf8') {
       $self->{skiprex} = '';
-      $self->{tokenrex} = qr/([\xF0-\xF4][\x80-\xBF][\x80-\xBF][\x80-\xBF]
-                             |[\xE0-\xEF][\x80-\xBF][\x80-\xBF]
-                             |[\xC2-\xDF][\x80-\xBF]
-                             |[\x00-\xFF])/x;
+      $self->{tokenrex} =
+           qr/([\xF0-\xF4][\x80-\xBF][\x80-\xBF][\x80-\xBF]
+              |[\xE0-\xEF][\x80-\xBF][\x80-\xBF]
+              |[\xC2-\xDF][\x80-\xBF]
+              |[\x00-\xFF])/x;
       $self->{processtoken} = '';
   }
   elsif ($params{type} eq 'word') {
       $self->{skiprex} = qr/[^a-zA-Z0-9]+/;
       $self->{skipinsert} = ' ';
-      $self->{tokenrex} = qr/([a-zA-Z]+|(\d+(\.\d+)?|\d*\.\d+)([eE][-+]?\d+)?)/;
-      $self->{processtoken} = sub { s/(\d+(\.\d+)?|\d*\.\d+)([eE][-+]?\d+)?/<NUMBER>/ }
+      $self->{tokenrex} =
+        qr/([a-zA-Z]+|(\d+(\.\d+)?|\d*\.\d+)([eE][-+]?\d+)?)/;
+      $self->{processtoken} = sub
+        { s/(\d+(\.\d+)?|\d*\.\d+)([eE][-+]?\d+)?/<NUMBER>/ }
   }
 
 =back
@@ -791,8 +801,9 @@ foolowing way:
 =head2 feed_tokens ( list of tokens )
 
   $ng3->feed_tokens('a');
+  $ng3->feed_tokens('b', 'c');
 
-This function manually supplies tokens.
+This function supplies tokens directly.
 
 =head2 process_text ( list of strings )
 
@@ -803,24 +814,43 @@ Process text, i.e., break each string into tokens and feed them.
 
 =head2 process_files ( file_names or file_handle_references)
 
+A usage example:
+
   $ng->process_files('somefile.txt');
 
-Process files, similarly to text.
-The files are processed line by line, so there should not be any
-multi-line tokens.
+This method is used to process one or more files, similarly to processing text.
+The files are processed line by line, so there should be no multi-line tokens.
+Instead of filenames we can also give as arguments file handle references
+when a file is already open.  In this way, we can use the standard input handle as in:
 
-=head2 get_ngrams(orderby=>'ngram|frequency|none',onlyfirst=>NUMBER,out=>filename|handle,normalize=>1)
+  $ng->process_files(\*STDIN);
+
+=head2 get_ngrams ( n => NUMBER, orderby => 'ngram|frequency|none', onlyfirst => NUMBER, out => filename|handle,normalize=>1)
 
 Returns an array of requested n-grams and their friequencies in order
 (ngram1, f1, ngram2, f2, ...).  The use of parameters is identical to
-the fuction C<to_string>.
+the function C<to_string>, except that the option 'spartan' is not applicable to C<get_ngrams> function.
 
-=head2 to_string(orderby=>'ngram|frequency|none',onlyfirst=>NUMBER,out=>filename|handle,normalize=>1,spartan=>1)
+Parameters:
+
+=over 4
+
+=item C<n>
+
+The parameter C<n> specifies the size of n-grams being retrieved.  The default value
+is the C<windowsize> field.  It should be less or equal than C<windowsize>.
+
+=back
+
+=head2 to_string ( orderby => 'ngram|frequency|none', onlyfirst => NUMBER, out => filename|handle, normalize => 1, spartan => 1 )
+
+Some examples:
 
   print $ng3->to_string;
   print $ng->to_string( orderby=>'frequency' );
   print $ng->to_string( orderby=>'frequency', onlyfirst=>10000 );
-  print $ng->to_string( orderby=>'frequency', onlyfirst=>10000, normalize=>1 );
+  print $ng->to_string( orderby=>'frequency', onlyfirst=>10000,
+                        normalize=>1 );
 
 Produce string representation of the n-gram tables.
 
@@ -865,6 +895,10 @@ is printed.
 
 =head2 encode_S ( string )
 
+This function translates any string in a /^\S*$/ compliant representation.
+It is primarely used in n-grams string representation to prevent white-space
+characters to invalidate the output format.  A usage example is:
+
   $e = Text::Ngrams::encode_S( $s );
 
 or simply
@@ -872,11 +906,14 @@ or simply
   $e = encode_S($s);
 
 if encode_S is imported.  Encodes arbitrary string into an \S* form.
-See F<http://www.cs.dal.ca/~vlado/srcperl/snip/encode_S> for detailed
+
+See F<http://web.cs.dal.ca/~vlado/srcperl/snip/encode_S> for detailed
 explanation.
 
 =head2 decode_S ( string )
 
+This is the inverse funcation of C<encode_S>.  A usage example is:
+
   $e = Text::Ngrams::decode_S( $s );
 
 or simply
@@ -884,37 +921,10 @@ or simply
   $e = decode_S($s);
 
 if decode_S is imported.  Decodes a string encoded in the \S* form.
+
 See F<http://www.cs.dal.ca/~vlado/srcperl/snip/encode_S> for detailed
 explanation.
 
-
-=head1 HISTORY AND RELATED WORK
-
-This code originated in my "monkeys and rhinos" project in 2000, and
-is related to authorship attribution project.  Some of the similar
-projects are (URLs can be found at my site):
-
-=over 4
-
-=item Ngram Statistics Package in Perl, by T. Pedersen at al. 
-
-This is a package that includes a script for word n-grams.
-
-=item Text::Ngram Perl Package by Simon Cozens
-
-This is a package similar to Text::Ngrams for character n-grams.
-As an XS implementation it is supposed to be very efficient.
-
-=item Perl script ngram.pl by Jarkko Hietaniemi
-
-This is a script for analyzing character n-grams.
-
-=item Waterloo Statistical N-Gram Language Modeling Toolkit, in C++ by Fuchun Peng 
-
-A n-gram language modeling package written in C++.
-
-=back
-
 =head1 PERFORMANCE
 
 The preformance can vary a lot depending on the type of file, in
@@ -934,7 +944,8 @@ the whole file.  However, after 4 minutes I gave up on waiting for
 n-grams to be printed.  The bottleneck seems to be encode_S function,
 so after:
 
-  ngrams.pl -n=10 --type=byte --orderby=frequency --onlyfirst=5000 1Mfile
+  ngrams.pl -n=10 --type=byte --orderby=frequency --onlyfirst=5000
+            1Mfile
 
 it took about 3:24 + 5 =~ 9 minutes to print.  After changing
 C<ngrams.pl> so that it provides parameter C<out> to C<to_string> in
@@ -943,18 +954,11 @@ module C<Ngrams.pm> (see Text::Ngrams), it still took:
 
 =head1 LIMITATIONS
 
-If a user customizes a type, it is possible that a resulting n-gram
-will be ambiguous.  In this way, to different n-grams may be counted
-as one.  With predefined types of n-grams, this should not happen.
-For example, if a user chooses that a token can contain a space, and
-uses space as an n-gram separator, then a trigram like this "x x x x"
-is ambiguous.
-
-Method process_file does not handle multi-line tokens by default.
+The method C<process_file> does not handle multi-line tokens by default.
 This can be fixed, but it does not seem to be worth the code
 complication.  There are various ways around this if one really needs
 such tokens:  One way is to preprocess them.  Another way is to read
-as much text as necessary at a time then to use process_text, which
+as much text as necessary at a time then to use C<process_text>, which
 does handle multi-line tokens.
 
 =head1 THANKS
@@ -978,26 +982,71 @@ that you used the module.
 
 Author:
 
- 2003-2012 Vlado Keselj http://web.cs.dal.ca/~vlado
+ 2003-2017 Vlado Keselj http://web.cs.dal.ca/~vlado
 
 Contributors:
-      2005 Chris Jordan (contributed get_ngrams)
+
+      2005 Chris Jordan (contributed initial get_ngrams method)
       2012 Magdalena Jankowska (utf8_character ngrams type)
 
 This module is provided "as is" without expressed or implied warranty.
 This is free software; you can redistribute it and/or modify it under
 the same terms as Perl itself.
 
+To acknowledge the use of this module in academic publications, please use
+a reference to the following paper:
+
+N-gram-based Author Profiles for Authorship Attribution.
+Vlado Keselj, Fuchun Peng, Nick Cercone, and Calvin Thomas. In Proceedings of
+the Conference Pacific Association for Computational Linguistics, PACLING'03,
+Dalhousie University, Halifax, Nova Scotia, Canada, pp. 255-264, August 2003. 
+http://web.cs.dal.ca/~vlado/papers/meta/Kes03.html
+
 The latest version can be found at F<http://web.cs.dal.ca/~vlado/srcperl/>.
 
+=head1 HISTORY
+
+This code originated in my "monkeys and rhinos" project in 2000, and
+is related to authorship attribution project.  After our papers on authorship
+attribution it was reformatted as a Perl module in 2003.
+
 =head1 SEE ALSO
 
-Ngram Statistics Package in Perl, by T. Pedersen at al.,
-Waterloo Statistical N-Gram Language Modeling Toolkit in C++ by Fuchun Peng,
-Perl script ngram.pl by Jarkko Hietaniemi,
-Simon Cozen's Text::Ngram module in CPAN.
+Some of the similiar projects and related resources are the following:
 
-The links should be available at F<http://web.cs.dal.ca/~vlado/nlp>.
+=over 4
+
+=item Ngram Statistics Package in Perl, by T. Pedersen at al. 
+
+This is a package that includes a script for word n-grams.
+
+=item Text::Ngram Perl Package by Simon Cozens
+
+This is another CPAN package similar to Text::Ngrams for character n-grams.
+As an XS implementation it is supposed to be very efficient.
+
+=item Perl script ngram.pl by Jarkko Hietaniemi
+
+This is a script for analyzing character n-grams.
+
+=item Waterloo Statistical N-Gram Language Modeling Toolkit, in C++ by Fuchun Peng 
+
+A n-gram language modeling package written in C++.
+
+=item CPAN N-gram module comparison article by Ben Bullock.
+
+The page is available
+  at F<http://www.lemoda.net/perl/cpan-n-gram-modules/> gives an interesting list of
+  modules, although the review seem to be superficial and only partially correct.
+  The following modules are listed in this review:
+  Algorithm::NGram, IDS::Algorithm::Ngram, Lingua::EN::Bigram, Linuga::EN::Ngram,
+  Lingua::Gram, Lingua::Identify, Text::Mining::Algorithm::Ngram,
+  Text::Ngram, Text::Ngram::LanguageDetermine, Text::Ngramize, Ntext::Ngrams, and
+  Text::Positional::Ngram.
+
+=back
+
+Some links to these resources should be available at F<http://web.cs.dal.ca/~vlado/nlp>.
 
 =cut
-# $Id: Ngrams.pm 395 2012-12-13 03:49:08Z vlado $
+# $Id: $
